/*
 * CARNAC interface
 * prediction of secondary structure of RNA
 *
 */

import java.io.*;
import java.util.*;
import java.awt.*;
import java.awt.geom.*;
import java.awt.event.*;
import javax.swing.*;
import javax.swing.event.*;

/**
 * This is the graphics Panel used for displaying the backbone representation of the secondary structure
 *
 *
 */
public class GraphicsPanel extends JPanel implements MouseListener{

    /* fonts */
    private static final Font defaultFont = new Font("Monospaced", Font.BOLD, 14);
    private static final Font nameFont = new Font("Courier", Font.BOLD, 14);
	private static final Font graduationFont = new Font("Courier", Font.PLAIN, 10);
	private static final Font infoFont = new Font("Courier", Font.BOLD, 12);
    private Font font = defaultFont;
	
	/* colors */
    private Color selectionColor = Color.blue;
    private Color noSelectionColor = Color.gray;
    private Color interStemColor = Color.lightGray;

    private static final int stemBoxHeight = 10;
	private int yAxisStructure = 90;
	private int yAxis;
    private static final int borderWidth = 20;
	private static final int structureHeight = 55;
	private static final int graphicHeight = structureHeight + 40;
	
    private int charWidth;
    private int charHeight = 12;

    private boolean empty = true;
    private boolean zoomMax = false;
    private boolean zoomMin = false;
    private boolean zoomSelection = false;
	private boolean viewSequence = false;
    private boolean selectionMode = false;
    private boolean invert = false;

    private int xMinVisible;
    private int xMaxVisible;
	
	private Rectangle visibleRectNoZoomSelection;
	private Rectangle visibleRect;

    private float scale =  1f;
	private float scaleNoZoomSelection = 1f;
    
    private int graphicWidth = 0;
    private int graphicWidthNoSelection = 0;
    private int graphicWidthSelection = 0; 

	private Structure structure;
	private ArrayList stemList;
    private Stem selectedStem;
    private Sequence sequence;

    private Graphics2D g2;

	private StructureFrame structureFrame; // link

    /**
     * Construct the graphic Panel
     */
    public GraphicsPanel(StructureFrame structureFrame) {
			setBackground(Color.white);
			this.structureFrame = structureFrame;
    }

	/**
	 * set the structure
	 */
	public void set(Structure structure) {
		empty = false;
		addMouseListener(this);
		this.structure = structure;
		this.sequence = structure.getSequence();
		this.stemList = structure.getStemList();
		graphicWidthNoSelection = sequence.length();
		zoomMax = false;
		selectionMode = false;
		//showAll();
		
	}
	

    /**
     * Mouse events
     */
    public void mousePressed(MouseEvent e) {
    
		selectedStem = getSelectedStem(e.getX(), e.getY());
	
		if ( selectedStem != null) { 
			selectionMode = true;
			zoomSelection = false;
			updateForSelection();
			repaint(getVisibleRect());
		} else { /* */
			selectionMode = false;
			zoomSelection = false;
			structureFrame.toggleZoomSelection();
			repaint(getVisibleRect());
		}
    }
	
    public void mouseEntered(MouseEvent e){}
    public void mouseExited(MouseEvent e){}
	public void mouseClicked(MouseEvent e){}
    public void mouseReleased(MouseEvent e){}
	
    private Stem getSelectedStem(int x, int y){
    
		Stem stem;

		for (int i=0; i < stemList.size() ; i++) {
			stem = (Stem) stemList.get(i);
			if (isSelected(stem, x, y))
				return stem;
		}
		return null;
    }
	
    private boolean isSelected(Stem stem, int x, int y) {
		stem = shift(stem);
		
		if (stem.x4  < xMinVisible  || stem.x1 > xMaxVisible)
			return false;
			
		if (stemSelection(stem.x1, stem.x2).contains(x, y) || stemSelection(stem.x3, stem.x4).contains(x, y))
			return true;
		else
			return false;
    }

    /*
     *
     *
     * Methods to Control
     *
     *
     */
    public void changeFont(Font font) {
		this.font = font;
		update(g2);
		if (zoomMax)
			setScale(charWidth);
		else
			setScale(this.scale);
	}
	
    /**
     * unselect All
     */
    public void selectNone(){
		if (selectionMode){
			zoomSelection = false;
            selectionMode = false;
			setScale(this.scale);
		}
	}
	
	/**
	 * Select next Stem
	 */
	public void selectNext() {
		if (!selectionMode) {
			selectionMode = true;
			selectedStem = (Stem) stemList.get(0);
			repaint(getVisibleRect());
			return;
		}
		for (int i=0; i < stemList.size() ; i++) {
			if ((Stem) stemList.get(i) == selectedStem){
				selectedStem = (Stem) stemList.get((i+1) % stemList.size());
				updateForSelection();
				if (zoomSelection)
					zoomMax();
				else
					repaint(getVisibleRect());
				return;
			}
		}
	}
	/**
	 * Select Previous Stem
	 */
	public void selectPrevious() {
		if (!selectionMode) {
			selectionMode = true;
			selectedStem = (Stem) stemList.get(stemList.size() - 1);
			repaint(getVisibleRect());
			return;
		}
		for (int i=0; i < stemList.size() ; i++) {
			if ((Stem) stemList.get(i) == selectedStem){
				if (i==0)
					i = stemList.size() -1;
				else
					i--;
				selectedStem = (Stem) stemList.get(i);
				updateForSelection();
				if (zoomSelection)
					zoomMax();
				else
					repaint(getVisibleRect());
				return;
			}
		}
	}
	
	/**
	 * Swap Sequence View (show/hide actgcgt...)
	 */
    public void swapViewSequence(){
		viewSequence = !viewSequence;
		if (viewSequence && !zoomMax)
			zoomMax();

		repaint(getVisibleRect());
    }
		
    public boolean viewSequence() {
		return viewSequence;
    }

    /**
     * Move Visible X translation
     */
    public void moveVisible(int shift){
		visibleRect = getVisibleRect();
		if (visibleRect.getX() <= 0 && shift < 0)
			return;
		if ( (visibleRect.getX() + visibleRect.getWidth()) >= (graphicWidth * scale + 2 * borderWidth - 1) && shift > 0)
			return;
		visibleRect.translate(shift, 0);
		scrollRectToVisible(visibleRect);
    }

	/**
     *  zoom to selection enabled
     */
	
	public boolean zoomSelection() {
		return zoomSelection;
	}
	
	boolean unZoomSelection = false;
	
    public void ToggleZoomToSelection(){
		if (!selectionMode)
			return;
		
		zoomSelection = !zoomSelection;
		
		if (! zoomSelection) {
			unZoomSelection = true;
			this.scale = this.scaleNoZoomSelection;
			setScale(this.scale);
			unZoomSelection = false;
		}else {
			this.scaleNoZoomSelection = this.scale;
			visibleRectNoZoomSelection = getVisibleRect();
			zoomMax();
		}

	}

    public void noZoomSelection(){
		zoomSelection = false;
		setScale(this.scale);
		structureFrame.toggleZoomSelection();
	}
	
	/**
     * Zoom to Max
     */
    public void zoomMax(){
        zoomMax = true;
        setScale((float) charWidth);
    }


	private Rectangle newRect(Rectangle oldRect, int center) {
		int halfWidth = (int) ((double) oldRect.getWidth() / 2d);
		return new Rectangle (center - halfWidth, (int) oldRect.getY(), (int) oldRect.getWidth(), (int) oldRect.getHeight());
	}

	private double center(Rectangle rect) {
		return rect.getWidth() / 2.0 + rect.getX();
	}
	
	private int xToR (int x) {
		return (int) (this.scale * x + borderWidth);
	}
	
	
    /**
     * set the Scale
     */
    public void setScale(float newScale){

			float oldScale = this.scale;

			setVisible(false);
			visibleRect = getVisibleRect();
			
			graphicWidth = graphicWidthNoSelection;
			if (zoomSelection && shiftForZoomSelection()) {
					graphicWidth = graphicWidthSelection;
			}
			float scaleMin = ((float) getVisibleRect().getWidth() - ((float) 2 * borderWidth))  / ((float) graphicWidth);
			if (newScale < scaleMin) {
				newScale = scaleMin;
				this.scale = scaleMin;
				zoomMin = true;
				zoomMax = false;
			}else if (newScale >= (float) charWidth) {
				zoomMax = true;
				zoomMin = false;
				newScale = charWidth;
				this.scale = charWidth;
			}else{
				viewSequence = false;
				zoomMax = false;
				zoomMin = false;
				this.scale = newScale;
			}

			int delta = (int) ((newScale / oldScale) * center(visibleRect) - center(visibleRect));
			
			Dimension dimension;
			dimension = new Dimension((int) (graphicWidthNoSelection * scale) + 2 * borderWidth, (int) visibleRect.getHeight());
			
			if (zoomSelection) {
					if (shiftForZoomSelection()) {
						updateForSelection();
						dimension.setSize(graphicWidthSelection * newScale + 2 * borderWidth, dimension.getHeight());
						delta = (int) (xToR(xHideMin) - center(visibleRect)) + 6; // 6 to adjust center
					}else{
						delta = (int) (xToR(selectedStem.getMiddle()) - center(visibleRect));
					}
			}
			if (unZoomSelection)
				visibleRect = visibleRectNoZoomSelection;
			else
				visibleRect.translate(delta, 0);

			if (getPreferredSize().getWidth() < dimension.getWidth()){
				setPreferredSize(dimension);
				scrollRectToVisible(visibleRect);
			}else{
				scrollRectToVisible(visibleRect);
				setPreferredSize(dimension);
			}
			repaint(getVisibleRect());
			setVisible(true);
    }
	
	/**
     * Show All
     */
    public void showAll(){
			setVisible(false);
			//selectionMode = false;
			zoomMax = false;
			zoomMin = false;
			viewSequence = false;
			graphicWidth  = graphicWidthNoSelection;
			scale = ((float) getVisibleRect().getWidth() - ((float) 2 * borderWidth))  / ((float) graphicWidth);
			setPreferredSize(new Dimension((int) getVisibleRect().getWidth(), getHeight()));
			structureFrame.toggleZoomSelection();
			repaint(getVisibleRect());
			setVisible(true);
    }
	

    /**
     * get the Scale
     */
    public float getScale(){
			return this.scale;
    }
	
	public boolean isZoomMin(){
		return zoomMin;
    }
	
	public boolean isZoomMax() {
		return zoomMax;
	}

	public boolean IsViewSequence() {
		if (viewSequence && zoomMax)
			return true;
		else
			return false;
	}
	


	/*
	 *
	 *  Privmites for Drawing
	 *
	 *
	 */

    /**
     * create an Arc
     */
	 
	private float arcParam = 0.7f;
    private Arc2D arc(int openMin, int openMax, int closeMin, int closeMax){
		//int stemWidth = (int) (scale * (float) (openMax - openMin + 1));
		//int stop = (int) ((scale) * (float) (closeMin + closeMax) / 2.0);
		//int start = (int)  ((scale * (float) (openMin + openMax) / 2.0));
		int start  = (int) (scale * (float) (openMax + 1));
		int stop   = (int) (scale * (float) closeMin);
		//double i = java.lang.Math.sqrt( (closeMin - openMax) );
		//double max = java.lang.Math.sqrt(graphicWidthNoSelection);
		double i = java.lang.Math.pow( (closeMin - openMax), arcParam );
		double max = java.lang.Math.pow(graphicWidthNoSelection, arcParam);
		int height = (int)  ( (structureHeight ) * ( i / max));
		return new Arc2D.Double( start + borderWidth, yAxis - height, stop - start,  2 * height, 0,180, Arc2D.OPEN);
    }
	 
	
    /**
     * Create a stem rectangle
     */
    private Rectangle2D stem(int xMin,int xMax){
		int width = (int) (scale * (float) (xMax + 1)) - (int) (scale * (float) (xMin));
		return new Rectangle2D.Double((int) (scale * (float) xMin) + borderWidth, yAxis, width, stemBoxHeight);
    }
		
    /**
     * Create a selection stem rectangle
     */
    private  Rectangle2D stemSelection(int xMin,int xMax){
		int width = (int) (scale * (float) (xMax + 1)) - (int) (scale * (float) (xMin));
		return new Rectangle2D.Double((int) (scale * (float) xMin) + borderWidth - 2, yAxis - 4 , width + 4, stemBoxHeight + 8);
    }		
	

	/*
	 *
	 * Shift for fit selection
	 *
	 */
	private boolean shiftForZoomSelection(){
		if (!selectionMode || !zoomSelection)
			return false;
		int selectionWidth = (int) (selectedStem.getWidth() * scale);
		if (selectionWidth >= getVisibleRect().getWidth() ) {
			return true;
		}
		else
			return false;
	}
    
    private int xHideMin, xHideMax, xShift;

    private void updateForSelection(){
		if (!selectionMode)
			return;
		int hideWidth = (int) (( (getVisibleRect().getWidth() - 2 * borderWidth) / 2.0 ) / scale);
		xHideMin =  selectedStem.x1 + hideWidth;
		xHideMax =  selectedStem.x4 - hideWidth;
		xShift = xHideMax - xHideMin;
		if (xShift > 0)
			graphicWidthSelection = graphicWidthNoSelection - xShift;
		else
			graphicWidthSelection = graphicWidthNoSelection;
	}

    
    /* Shift the x if selection return True if the stem is visible */
    private Stem shift(Stem stem){

		if (!zoomSelection)
			return stem;
			
		Stem shiftedStem = new Stem(stem);
        if (shiftForZoomSelection()){
            if ( stem.x4 < xHideMin || stem.x1 > xHideMax || (stem.x2 < xHideMin && stem.x3 > xHideMax)) {
                if (stem.x3 > xHideMax) { // shift the close part of the Stem

                    shiftedStem.x3 = stem.x3 - xShift;
                    shiftedStem.x4 = stem.x4 - xShift;
                          
                }
                if (stem.x1 > xHideMax) { // shift all the Stem
                    shiftedStem.x1 = stem.x1 - xShift;
                    shiftedStem.x2 = stem.x2 - xShift;
                    shiftedStem.x3 = stem.x3 - xShift;
                    shiftedStem.x4 = stem.x4 - xShift;
                 }
             }else{ // not visible
				 
                 return new Stem(); // an empty Stem
             }
        }
		
        return shiftedStem;
    }

    /**
     * Draw a stem corresponding to stem data
     */
    private void drawStem (Stem stem){
	
		Stem shiftedStem = shift(stem);
                
        if (shiftedStem.isEmpty())
            return;

		if (shiftedStem.x4  < xMinVisible  || shiftedStem.x1 > xMaxVisible)
			return;
	
		yAxis = yAxisStructure;
		/*
         * Stem part
         */
		 
		Color color = g2.getColor();
        g2.setColor(new Color(225,225,225));  
        g2.fill(stem(shiftedStem.x1, shiftedStem.x2));
        g2.fill(stem(shiftedStem.x3, shiftedStem.x4));
		g2.setColor(color); 
		 
		 
        g2.setStroke(new BasicStroke(1.5f));
        g2.draw(stem(shiftedStem.x1, shiftedStem.x2));
        g2.draw(stem(shiftedStem.x3 ,shiftedStem.x4));
        

        /*
         * Arc
         */
		g2.draw(arc(shiftedStem.x1, shiftedStem.x2, shiftedStem.x3, shiftedStem.x4));

    }
	
	
	private void drawSequence(Sequence sequence){
        int yShift = 5;
        yAxis = yAxisStructure;
		
		if (shiftForZoomSelection() && zoomSelection) {
			if (xHideMin -3 > xMinVisible)
				g2.drawString(sequence.sub(xMinVisible, xHideMin - 3), borderWidth + scale * xMinVisible, yAxis + yShift + charHeight);

			if (xMaxVisible + (xHideMax - xHideMin) > xHideMax + 3) {
				g2.drawString(sequence.sub(xHideMax + 3, xMaxVisible + (xHideMax - xHideMin)), borderWidth + scale * (xHideMin + 3), yAxis + yShift + charHeight);
			}
			Color currentColor = g2.getColor();
			g2.setColor(Color.red);
			g2.drawString("......", borderWidth + scale * (xHideMin -3), yAxis + yShift + charHeight);
			g2.setColor(currentColor);
			return;

        } 
		g2.drawString(sequence.sub(xMinVisible, xMaxVisible ), borderWidth + scale * xMinVisible, yAxis + yShift + charHeight);
    }
	
    private void drawStemWithSequence(Sequence sequence, Stem stem, boolean left){
	
		Stem shiftedStem = shift(stem);
	
		if (shiftedStem.isEmpty())
			return;

		if (shiftedStem.x4 < xMinVisible || shiftedStem.x1 > xMaxVisible)
			return;

        int yShift = 5;
        yAxis = yAxisStructure;

		if (left)
            g2.drawString(sequence.sub(shiftedStem.i1, shiftedStem.i2 + 1), shiftedStem.x1 * scale + borderWidth, yAxis + yShift + charHeight);
        else
            g2.drawString(sequence.sub(shiftedStem.i3, shiftedStem.i4 + 1), shiftedStem.x3 * scale + borderWidth, yAxis + yShift + charHeight);
    }

    /**
     * Draw a stem and data associated
     */
    private void drawStemWithData(Sequence sequence, Stem stem){

		int yShift;
		int yShift2;
		int yShift3;
		Color color = g2.getColor(); // save current color

		String strStem = sequence.sub(stem.i1, stem.i2 + 1) + "    " + sequence.sub(stem.i3, stem.i4 + 1);

		
		yShift = 5;
		yAxis = yAxisStructure;

		g2.setStroke(new BasicStroke());
		g2.setColor(Color.blue);

		int startPos = (int) (stem.x1 * scale);
		if (zoomMax){
			drawStemWithSequence(sequence, stem, false);
			drawStemWithSequence(sequence, stem, true);
		}else{
			if (stem.x1 * scale + strStem.length() * charWidth > (graphicWidthNoSelection -4) * scale )
				startPos = max ((int) (stem.x4 *scale - strStem.length() * charWidth), 0);
			g2.drawString(strStem, startPos  + borderWidth, yAxis + yShift + charHeight);
		}
			/* data */
		yShift2 = 40;
		g2.setColor(Color.black);
		g2.setFont(infoFont);
		String strIdx = "[" + Integer.toString(stem.i1 + 1) + "." +  Integer.toString(stem.i2 + 1) +
		"]  [" + Integer.toString(stem.i3 + 1) + "." +  Integer.toString(stem.i4 + 1) + "]";
		g2.drawString(strIdx, startPos + borderWidth , yAxis + yShift2);
		
		g2.setColor(color); // restore color

    }
	
	private String space(int length) {
		StringBuffer buf = new StringBuffer();
		for (int i = 0; i < length; i++)
			buf.append(" ");
		return buf.toString();
	
	}

	private int min(int a, int b){
		if (a > b)
			return b;
		else
			return a;
	}

	private int max(int a, int b){
		if (a > b)
			return a;
		else
			return b;
	}

    private void setColor(Stem stem) {
		if (stem == selectedStem && selectionMode) {
			g2.setColor(selectionColor);
		}else{
			g2.setColor(noSelectionColor);
		}
	}

	private void explore() {
		Stem stem;
		g2.setColor(noSelectionColor);
		for (int i=0; i < stemList.size(); i++) {
			stem = (Stem) stemList.get(i);
			if (zoomMax && viewSequence) {
				drawStemWithSequence(sequence, stem, false);
				drawStemWithSequence(sequence, stem, true);
			}
				drawStem(stem);
		}
   }

	private int shift(int i) {
		if (selectionMode && zoomSelection && shiftForZoomSelection()) {
			if ( i < xHideMin)
				return i;
			else if (i > xHideMax)
				return i - xShift;
			else
				return -1;
		}else{
			return i;
		}
	}
	
    private int yGrad =  yAxisStructure + 50;
	
	private void drawVerticalGrad(int i) {
		int iShifted = shift(i);
		if (iShifted < xMinVisible - 10 || iShifted > xMaxVisible + 10 || iShifted < 0)
			return;
		
		String label = Integer.toString(i);
		double halfWidth = ((Rectangle2D) graduationFont.getStringBounds(label, g2.getFontRenderContext())).getWidth() / 2.0;
		g2.drawLine ((int) (borderWidth + iShifted * scale), yGrad, (int) (borderWidth + iShifted * scale), yGrad - 4);
		g2.drawString (label, (int) (borderWidth + iShifted * scale - halfWidth), yGrad + 10);
	}
	
	private void drawGraduation() {
		/* Graduation */

		if (selectionMode && zoomSelection && shiftForZoomSelection()){
			g2.drawLine (borderWidth, yGrad, (int) ((graphicWidthSelection) * scale) + borderWidth, yGrad);
		}else{
			g2.drawLine (borderWidth, yGrad, (int) (graphicWidth * scale) + borderWidth, yGrad);
		}
		
		g2.setFont(graduationFont);
		drawVerticalGrad(0);
		drawVerticalGrad(sequence.length());
		
		int step;
		
		if (scale > 5.0)
			step = 5;
		else if (scale > 3.0)
			step = 10;
		else if (scale > 2.0)
			step = 20;
		else if (scale > 1.0)
			step = 50;
		else
			step = 100;
			
		for (int i = step; i < sequence.length() - step; i += step)
			drawVerticalGrad(i);
	
	}
	
	/**
     * Draw the Secondary Structure
     */
	private void drawStructure(){
		g2.setFont(nameFont);
		g2.setColor(Color.darkGray);
		g2.setStroke(new BasicStroke(1.2f));

	    int yLine =  yAxisStructure + stemBoxHeight + 2;
	    int yName = 25;
		//g2.setRenderingHint(RenderingHints.KEY_TEXT_ANTIALIASING, RenderingHints.VALUE_TEXT_ANTIALIAS_ON);
		g2.drawString("[ " + sequence.getName() + " ]", 20, yName);
		//g2.setRenderingHint(RenderingHints.KEY_TEXT_ANTIALIASING, RenderingHints.VALUE_TEXT_ANTIALIAS_OFF);
		
		
		g2.setColor(Color.darkGray);
		if (selectionMode && zoomSelection && shiftForZoomSelection()){
			g2.drawLine (borderWidth, yLine, (int) ((graphicWidthSelection) * scale) + borderWidth, yLine);
		}else{
			g2.drawLine (borderWidth, yLine, (int) (graphicWidth * scale) + borderWidth, yLine);
		}
		
		drawGraduation();
		
		g2.setFont(font);
		if (zoomMax && viewSequence){
			g2.setColor(interStemColor);
			drawSequence(sequence);
		}
		if (selectionMode){
			
		}
		explore();
		if (selectionMode){
			g2.setColor(selectionColor);
			if (zoomMax){
				drawStemWithSequence(sequence, selectedStem, false);
				drawStemWithSequence(sequence, selectedStem, true);
			}
			drawStemWithData (sequence, selectedStem);
			drawStem(selectedStem);
		}
        
	}

	public void paintComponent(Graphics g) {
		
		g2 = (Graphics2D) g;
		super.paintComponent(g);
	
		if (empty)
			return;

	    /*
	     * configure the Graphic
	     */
		g2.setRenderingHint(RenderingHints.KEY_ANTIALIASING, RenderingHints.VALUE_ANTIALIAS_ON);
		//g2.setRenderingHint(RenderingHints.KEY_TEXT_ANTIALIASING, RenderingHints.VALUE_TEXT_ANTIALIAS_OFF);
		g2.setRenderingHint(RenderingHints.KEY_RENDERING, RenderingHints.VALUE_RENDER_SPEED);


		visibleRect = getVisibleRect();
		int aMinVisible = (int) visibleRect.getX();
		int aMaxVisible = aMinVisible + (int) visibleRect.getWidth();
		
		xMinVisible = max ((int) ((float) (aMinVisible - borderWidth) / this.scale) - 1, 0);
		xMaxVisible = min ((int) ((float) (aMaxVisible - borderWidth) / this.scale) + 1, graphicWidthNoSelection);
		
		charWidth = (int) ((Rectangle2D) font.getStringBounds("i", g2.getFontRenderContext())).getWidth();
		charHeight = (int) ((Rectangle2D) font.getStringBounds("i", g2.getFontRenderContext())).getHeight();
		drawStructure();


    }
}
