
/*-----------------------------------------------------------*/
/* main.c : carnac                                           */
/*-----------------------------------------------------------*/

#include <stdio.h>
#include <unistd.h>
#include <errno.h>
#include <string.h>
#include <stdlib.h>
#include <time.h>
#include "parse.h"            // parsing of config file
#include "types.h"            // (allmost) all types declared
#include "get_sequences.h"    // get sequences in 'fasta' format
#include "find_blocks.h"      // find anchor points
#include "print_alignment.h"  // print anchor points
#include "find_stems.h"       // find potential stems
#include "correlate.h"        // filter stems
#include "fold.h"             // dynamic cofolding
#include "tree.h"             // tree library
#include "print_folding.h"    // print pairwise (alignment) foldings
#include "accumulation.h"     // gather results
#include "graph.h"            // graph library
#include "clique.h"           // create stem graphs
#include "nuss.h"             //

// -- all --
string_t infile_arg;          // config file
string_t infile_seq;          // sequence file (FASTA)
string_t infile_stem;         // stem file (RNAML)
string_t infile_forced;       // forced stems file (RNAML)
string_t refname;             // reference filename
int nb_seq;                   // nb of sequences
sequence_list_t all_seq;      // list of all the sequences
stem_list_t *all_stems;       // list of the full stem lists
stem_list_t *forced_stems;    // list of forced stems lists
stem_list_t *R;               // folded stems
stem_list_t *finalR;          // final folded stems
stem_list_t *remain;          // remaining compatible stems
block_list_t **block_list;    // list of all blocks
matrix2_t similarity;         // percentage of similarity (prim level)

// -- current --
int i, j;                     // index of the 2 current sequences
sequence_list_t seq;          // the 2 current sequences
stem_list_t all0, all1;       // lists of current potential stems
stem_list_t K0, K1;           // lists of current filtered stems
stem_list_t L0, L1;           // same lists ordered by end of stems
stem_list_t S0, S1;           // lists of folded stems
block_list_t block;           // current list of blocks
matrix2_t correlated;         // correlation of stems
node_t *stree;                // cofolding tree
unsigned long clock_b, clock_e, tt;    // CPU time
int brk_lp = 0;               // bool: reach fast end of the loop

int main (int argc, char **argv) {

  // clock
  clock_b = (unsigned long) clock () / 1000000;

  // reading command line & config file
  parse_command_line (argc, argv);
  parse_config_file (infile_arg);

  // reading sequences
  all_seq = get_sequences (infile_seq);

  // discard too similar sequences
  all_seq = reduce_set (all_seq);
  free_matrix2 (similarity, nb_seq);
  print_infos ();

  // finding potential stems
  if (infile_forced) get_forced_stems (infile_forced);
  if (infile_stem) get_stems (infile_stem); else find_stems ();
  if (infile_forced) filter_stems ();

  // reading structure
  get_true_structure ();

  // print all stems in 'stems.out'
  // for (i=0; i<nb_seq; i++) print_stem_list (all_stems[i], i);

  // pairwise confrontation
  for (i=0; i<nb_seq; i++) {
    for (j=i+1; j<nb_seq; j++) {
      // clock
      clock_b = (unsigned long) clock () / 1000000;
      // select 2 sequences
      tournament (i, j, all_stems, &all0, &all1);
      // find & print anchor points
      block_list[i][j] = find_blocks (seq);
      block = block_list[i][j];
      if (break_loop(i,j,"no block")) goto end_loop;
      if (OUTPUT) print_blocks (block_list, i, j);
      if (OUTPUT) print_alignment (seq, block_list[i][j], i, j);
      // filter stems (stems/stems/stems/stems/uncovary)
      correlated = pre_select (&K0, &K1, &L0, &L1, !COVAR);
      if (break_loop(i,j,"no stem or too many stems")) goto end_loop;
      // cofolding (stems/stems/verbose/shorten)
      stree = fold (&S0, &S1, 1, 0);
      if (break_loop(i,j,"matrix too large")) goto end_loop;
      retrieve_folding (K0, K1, stree);
      free_matrix2 (correlated, stem_list_length (K0));
      if (OUTPUT) print_folding (seq, i, j);
      // print co-foldings
      if (OUTPUT) {
	// pairwise 'connect' files
	create_str_ct_file (i, j, 0, K0, seq, stree);
	create_str_ct_file (j, i, 1, K1, seq, stree);
	// pairwise '.out' file: the list of co-folded stems
	create_str_out_file (i, j, K0, K1, seq, stree);
	// pairwise '.str' file: co-tree in 'txt' format
	create_str_str_file (i, j, K0, K1, seq, stree);
      }

      // gather results
      cat_stems (S0, i);
      cat_stems (S1, j);

      // clock
      clock_e = (unsigned long) clock () / 1000000;
      tt = clock_e - clock_b;
      if (VERBOSE) printf (" (%.2lu:%.2lu:%.2lu)\n", tt / 3600, (tt%3600/60), tt%60);

      end_loop :;

    }
  }

  // compute stem frequencies
  conglomerate_stems (all_seq);

  // build stem graph
  build_graphs ();

  // nussinov with index of stems (seq/stems/shorten)
  //  for (i=0; i<nb_seq; i++) stree = nuss (i, R[i], 1);

  // compute final structure (seq/magnify_energy/thd1/thd2)
  final_structure (all_seq, 0, 30, 40);

  // compare with known structure
  if (COMPARE) final_compare ();

  // retrieve remaining stems
  get_remaining_stems ();
  if (VERBOSE) printf("\n");

  // generate latex file for viewing structures
  if (LATEX) generate_latex(argc, argv);

  // free memory
  free (all_stems);
  free (finalR);
  free (R);

  // clock
  tt = (unsigned long) clock () / 1000000;
  if (VERBOSE) printf ("\n  overall time : %.2lu:%.2lu:%.2lu\n\n",
		       tt / 3600, (tt%3600/60), tt%60);
  return 0;
}

